{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "6b459ebf",
   "metadata": {},
   "source": [
    "![_](pancake_sort.png)\n",
    "\n",
    "# The Pancake Problem\n",
    "This is a specialized sorting problem, where we intentionally limit our sorting capabilities. Imagine a stack of pancakes of varying sizes. We want to sort our stack so that the largest pancake is on the bottom, and that they decrease in size as the stack moves upwards. However, the only operation we can perform, is to insert a spatula between two pancakes in the stack, and then flip *all* pancakes above that one around in one go.\n",
    "\n",
    "The [Pancake Problem](https://en.wikipedia.org/wiki/Pancake_sorting) is finding a general algorithm for how to sort such a pancake stack in the least amount of flips. It is an unsolved problem, and it is unknown what the optimal solution is.\n",
    "\n",
    "Here we first create a class `PancakeStack` that extends the built-in `list`-class. We add a `.flip(n)` method that flips the top $n$ pancakes (i.e., the first $n$ pancakes). This method also automatically keeps track of the number of flips that have been used in the `nr_of_flips`-attribute. Finally we also add a boolean property `is_sorted`, that we can use to check if we are finished with sorting.\n",
    "\n",
    "\n",
    "```python\n",
    "class PancakeStack(list):\n",
    "    \"\"\"A specialized list class, implementing the flip action.\n",
    "\n",
    "    A pancake stack is represented from top to bottom.\n",
    "    This class is intended for working on the Pancake Problem.\n",
    "    Ref: https://en.wikipedia.org/wiki/Pancake_sorting\n",
    "    \"\"\"\n",
    "\n",
    "    def __init__(self, *args, **kwargs):\n",
    "        super().__init__(*args, **kwargs)\n",
    "        self.nr_of_flips = 0\n",
    "\n",
    "    def flip(self, n):\n",
    "        \"\"\"Flip the top n pancakes of the stack.\"\"\"\n",
    "        if n < 1:\n",
    "            raise ValueError(\"Must flip at least one pancake!\")\n",
    "        elif n > len(self):\n",
    "            raise ValueError(\n",
    "                f\"Asked to flip {n} pancakes, but stack only has {len(self)}\"\n",
    "            )\n",
    "\n",
    "        self[:n] = reversed(self[:n])\n",
    "        self.nr_of_flips += 1\n",
    "\n",
    "    @property\n",
    "    def is_sorted(self):\n",
    "        return self == sorted(self)\n",
    "\n",
    "    def sort(self):\n",
    "        raise NotImplementedError\n",
    "```\n",
    "\n",
    "## Exercise 1a) Implementing the pancake sort itself\n",
    "\n",
    "Make either a stand-alone function, `pancake_sort`, that takes in a `PancakeStack` object as input **or** extend the `PancakeMethod` with a `PancakeStack.sort` method that sorts the stack.\n",
    "Either of these should sort the stack in an increasing order. The only way you are allowed to rearrange the stack is using the `PancakeStack.flip` method. You are not allowed to use any external libraries as aid (for example `np.argmax`, but you can define your own utility functions/methods as desired. (Hint: Defining an `argmin` or `argmax` function/method can be very helpful!).\n",
    "\n",
    "One of the following two test-snippets should work\n",
    "\n",
    "**As stand-alone function**\n",
    "***\n",
    "```python\n",
    "stack = PancakeStack([2, 1, 0, 3, 5, 4, 6])\n",
    "pancake_sort(stack)\n",
    "\n",
    "assert stack.is_sorted\n",
    "print(\"Sorted stack:\", stack)\n",
    "print(f\"{stack.nr_of_flips} flips were used in total.\")\n",
    "```\n",
    "***\n",
    "\n",
    "**As class method**\n",
    "```python\n",
    "stack = PancakeStack([2, 1, 0, 3, 5, 4, 6])\n",
    "stack.sort()\n",
    "\n",
    "assert stack.is_sorted\n",
    "print(\"Sorted stack:\", stack)\n",
    "print(f\"{stack.nr_of_flips} flips were used in total.\")\n",
    "```\n",
    "\n",
    "```python\n",
    "class PancakeStack(PancakeStack):\n",
    "    @staticmethod\n",
    "    def argmax(sequence):\n",
    "        assert sequence, \"empty list\"\n",
    "        max_index = 0\n",
    "        max_value = sequence[0]\n",
    "        for i, value in enumerate(sequence[1:]):\n",
    "            if value > max_value:\n",
    "                max_value = value\n",
    "                max_index = i + 1\n",
    "        return max_index\n",
    "\n",
    "    def sort(self):\n",
    "        \"\"\"Sort the Pancake stack using only flips\"\"\"\n",
    "        unsorted = len(stack)\n",
    "        while not stack.is_sorted:\n",
    "            # Find largest pancake in unsorted stack\n",
    "            biggest_remaining = self.argmax(self[:unsorted])\n",
    "            # If it is at the bottom already, do nothing\n",
    "            if biggest_remaining == unsorted - 1:\n",
    "                pass\n",
    "\n",
    "            # If it is at the top already, flip whole remaining stack\n",
    "            if biggest_remaining == 0:\n",
    "                stack.flip(unsorted)\n",
    "\n",
    "            # If it is somewhere in the middle, flip to top, then to bottom\n",
    "            else:\n",
    "                stack.flip(biggest_remaining + 1)\n",
    "                stack.flip(unsorted)\n",
    "\n",
    "            # Unsorted stack is now 1 shorter\n",
    "            unsorted -= 1\n",
    "```\n",
    "\n",
    "Running the test snippet now gives\n",
    "\n",
    "```python\n",
    "stack = PancakeStack([2, 1, 0, 3, 5, 4, 6])\n",
    "stack.sort()\n",
    "\n",
    "assert stack.is_sorted\n",
    "print(\"Sorted stack:\", stack)\n",
    "print(f\"{stack.nr_of_flips} flips were used in total.\")\n",
    "```\n",
    "\n",
    "The code passes our test!\n",
    "\n",
    "\n",
    "## Exercise 1b) Algorithm analysis\n",
    "\n",
    "How many flips did it take your function to sort the example stack given?\n",
    "\n",
    "Analyze your function. What is the best case number of flips for a stack of $n$ pancakes? What is the worst case?\n",
    "\n",
    "\n",
    "\n",
    "**Example solution:**\n",
    "\n",
    "* The best case will be a pre-sorted stack. Our code, as we have implemented it, will recognize such a stack, in which case no flips will be performed. The best case is therefore 0 flips.\n",
    "* The worst case occurs when the maximum number of flips occurs. In our code we sort one and one pancake to the bottom of the remaining stack. We use a maximum of 2 flips per pancake, and thus the worst case will be $2n$ flips. (Actually it will be slightly lower than this, see 1d).\n",
    "\n",
    "\n",
    "## Exercise 1c) Testing *all permutations* for smaller stacks\n",
    "\n",
    "We now want to test our estimates from (1b), and our implementation. To do this we simply want to try *all possible stacks* of various sizes (assuming no duplicate pancakes). To produce all possible permutations of a given stack size, we can use the function `itertools.permutations`, which lets us easily loop over all possible permutations (i.e., different possible configurations).\n",
    "\n",
    "However, do note that the number of possible permutations for a stack of size $n$ is $n!$, which grows very, very rapidly. We therefore limit ourselves to stack sizes up to 7.\n",
    "\n",
    "Using a loop, find the lowest, the highest and the average number of flips used for stacks of sizes 3, 4, 5, 6, and 7. Plot your results as the three curves in a flip-stack size plot.\n",
    "Find the worst case, the best case and the average case of all permutations of stacks of different sizes and plot your results as three curves in the same plot.\n",
    "\n",
    "```python\n",
    "import itertools\n",
    "import numpy as np\n",
    "import matplotlib.pyplot as plt\n",
    "\n",
    "stack_sizes = range(3, 8)\n",
    "worst = []\n",
    "best = []\n",
    "avg = []\n",
    "\n",
    "for n in stack_sizes:\n",
    "    flips = []\n",
    "    for stack in itertools.permutations(range(n)):\n",
    "        stack = PancakeStack(stack)\n",
    "        stack.sort()\n",
    "        flips.append(stack.nr_of_flips)\n",
    "\n",
    "    worst.append(max(flips))\n",
    "    best.append(min(flips))\n",
    "    avg.append(sum(flips) / len(flips))\n",
    "\n",
    "plt.plot(stack_sizes, worst, \"o-\", label=\"Worst case\")\n",
    "plt.plot(stack_sizes, best, \"o-\", label=\"Best case\")\n",
    "plt.plot(stack_sizes, avg, \"o-\", label=\"Avg case\")\n",
    "plt.xlabel(\"Stack size\")\n",
    "plt.ylabel(\"Number of flips used\")\n",
    "plt.legend()\n",
    "plt.show()\n",
    "```\n",
    "\n",
    "## Exercise 1d) Checking our analysis\n",
    "\n",
    "Compare your analysis in 1b with your actual results in 1c. Were your best and worst case estimates correct? What is your average case?\n",
    "\n",
    "\n",
    "**Example Solution**: First we can look at the curve for the best case, which we see is 0 for all stack sizes. This is because when we loop over *all* permutations, we also get the case with the perfectly sorted stack, which will require 0 flips. Our analysis is therefore confirmed, that the best case is 0 flips.\n",
    "\n",
    "Next, let us look at the best case. The curve does grow linearly with $n$, as we analyzed. However, we stated that it was $2n$, but the curve is actually lower than this. Printing out the actual numbers shows them to be\n",
    "\n",
    "\n",
    "\n",
    "```python\n",
    "print(worst)\n",
    "```\n",
    "\n",
    "This is because we forgot to compensate for the the fact that the final pancake in the stack never will require any flips, so we go from $2n$ to $2(n-1)$. Likewise, the second-to-last pancake cannot possibly require *two* flips to get to the bottom of the remaining stack, so we have $2(n-1)-1 = 2n -3$. Which is indeed what we found:\n",
    "\n",
    "```python\n",
    "print([2 * n - 3 for n in (3, 4, 5, 6, 7)])\n",
    "```\n",
    "\n",
    "So we see the actual worst case is $2n-3$. For larger $n$ than we use here, this modification is minor, and for larger stacks our worst case will approach $2n$.\n",
    "\n",
    "The average case we have not analyzed earlier, but we have found it in our code now\n",
    "\n",
    "```python\n",
    "print(avg)\n",
    "```\n",
    "\n",
    "Note that it is slightly below the worst case, but does scale linearly with $n$, as the worst case does, but perhaps with an ever so slightly lower incline. We can use linear regression to find a best-fit expression\n",
    "\n",
    "```python\n",
    "from scipy.stats import linregress\n",
    "\n",
    "print(linregress(stack_sizes, y=avg))\n",
    "```\n",
    "\n",
    "So the average case is roughly $1.78n -3.6$, or if we ignore the intercept and round it of a bit, we can write it out as simply $\\frac{9}{5}n$.\n",
    "\n",
    "\n",
    "## Exercise 1e) Checking if a few randomly selected larger stacks (i.e., random sampling)\n",
    "\n",
    "We now want to check some stacks of larger sizes, like 100, 1000 or 10000 pancakes. For such stacks, going through all possible permutations is completely infeasible. However, we can through a few *randomly selected* permutations and check those. Explain why this should give us a good estimate of the average run time, but not the best/worst cases?\n",
    "\n",
    "Estimate the average number of flips needed to sort stacks of 100, 1000 and 10000 pancakes by sorting respectively 100, 10, and 1 randomly shuffled stacks. You can for example use `random.shuffle` to shuffle the stacks.\n",
    "\n",
    "```python\n",
    "from random import shuffle\n",
    "\n",
    "trials = ((100, 100), (1000, 10), (10000, 1))\n",
    "avg_flips = []\n",
    "\n",
    "for size, nr_of_trials in trials:\n",
    "    flips = []\n",
    "    for i in range(nr_of_trials):\n",
    "        stack = PancakeStack(range(size))\n",
    "        shuffle(stack)\n",
    "        stack.sort()\n",
    "        flips.append(stack.nr_of_flips)\n",
    "\n",
    "    avg_flips.append(sum(flips) / len(flips))\n",
    "\n",
    "print(\"Stack size | Avg Flips\")\n",
    "for size, avg in zip([size for size, nr_of_trials in trials], avg_flips):\n",
    "    print(\"{:10} | {:10.0f}\".format(size, avg))\n",
    "```\n",
    "\n",
    "**Example Solution:** From this run, it seems the average number of flips for larger stacks lies closer to $2n$, than the $1.78n$ we found in (1d) i.e., it gets closer to the worst case.\n",
    "\n",
    "This actually makes a lot of sense if you think about it: The worst case occurs when we *always* use two flips to get the next biggest pancake to the bottom of the remaining stack. We need 2 flips if the biggest pancake isn't already at the bottom or top of the remaining stack. But what is that chance of this occurring in a randomly shuffled stack? The bigger the remaining stack is, the lower the odds of this occurring will be. For large stacks, we will therefore \"get lucky\" very rarely, and we will need to use 2 flips for the vast majority of pancakes. E.g., the average number of flips for a stack of a 100 pancakes came out at 193, when we know the worst case is $2\\cdot 100 - 3 = 197$, i.e., we only avoid a measly 4 flips on average.\n",
    "\n",
    "This might indicate that our solution might not be the approach for solving larger pancake stacks.\n",
    "\n",
    "For reference, the *actual* best solution is unknown, but it has been shown that the best possible *worst case* solution to the problem will need somewhere between $\\frac{15}{14}n$ and $\\frac{18}{11}n$ flips, which is quite a lot lower than our method's $2n$."
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "formats": "md:myst",
   "text_representation": {
    "extension": ".md",
    "format_name": "myst"
   }
  },
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "source_map": [
   11
  ]
 },
 "nbformat": 4,
 "nbformat_minor": 5
}